/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//

#include "vtl-calc.h"



using namespace std;

namespace ns3 {

Ptr<VtlCalc>   VtlCalc::m_vtlcalcobj;
Ptr<VtlCalc> 
VtlCalc::GetPtr()
{
  if (m_vtlcalcobj == NULL)
    m_vtlcalcobj = CreateObject<VtlCalc> ();
  return m_vtlcalcobj;
}

VtlCalc::VtlCalc () {
  p.stddev = 4.1;
  p.txPowerDbm = 20;
  p.rxThres = -92;
  p.roadsize = 10;
  p.stopdist = 4;
  p.offset = p.stopdist / 2.0;      // distance road middle to vehicle
  p.systemloss = 4.8;
  p.suburban = false;
  p.maxdist = 100;     // maximum observation distance (e.g. road length)
  p.prob = 0.99;
  p.pgr = 10;
  p.stddev = 4.1;
  p.maxspeed = 14;
  p.collprob = 0.0;
    
  p.lambda = 300000000.0 / 5.90e9;
  p.height = 1.5;
}

VtlCalc::~VtlCalc () {

}


double
VtlCalc::IsRoadEmpty(double curDist, double maxTime, Road::roadrelation o)
{
  double dist;
  double diststep;
  double probNeg=1;
  double timestep;
  double curtime=0;
  
  // calculate the probability that no packet has been received
  diststep = p.maxspeed / p.pgr;
  timestep = 1.0 / p.pgr;
  for (dist = curDist; dist < p.maxdist && curtime < maxTime; dist += diststep, curtime+=timestep) {
    probNeg = probNeg * (1.0-GetSingleRecieveProb(dist, o));
    if ((1.0 - probNeg) > p.prob)
      return true;
  }
  
  return false;
}


double
VtlCalc::GetSingleRecieveProb(double dist, Road::roadrelation o)
{
  double rxPwr;
  Vector senderPos = getSenderPosition(dist);
  Vector receiverPos = getReceiverPosition(dist, o);
  
  rxPwr = calcRxPwr(senderPos, receiverPos);
  
  // get probability that rxPwr > m_rxThres and no collision occured
  return calcRxProb(rxPwr) * (1.0-p.collprob);
}


inline double
VtlCalc::calcRxProb (double rxPwr)
{
  return Phi(rxPwr-p.rxThres, p.stddev);
}


inline Vector
VtlCalc::getSenderPosition (double dist)
{
  return Vector(dist+p.stopdist, p.offset, 0);
}

inline Vector
VtlCalc::getReceiverPosition (double dist, Road::roadrelation o)
{
  switch (o) {
  case Road::LOS:
    return Vector(-1.0*(dist+p.stopdist), -1.0 * p.offset, 0);
  case Road::RIGHT:
    return Vector(-1.0 * p.offset, dist+p.stopdist, 0);
  case Road::LEFT:
    return Vector(p.offset, -1.0 * (dist+p.stopdist), 0);
  default:
    NS_ASSERT(false);
    return Vector();
  }
  
}




double
VtlCalc::calcRxPwr(const Vector &srcPos, const Vector &dstPos)
{
  double nlosPowerDbm, rxPowerDbm;
  double dr, dt;
  double losPrDbm;
  double Pr0;
  double xt;
  bool los = false;
  
    
  // check if vehicles are in LOS
  // this checks if both vehicles are on the same road
  // or if sender or receiver is on intersection
  if (srcPos.x <= p.roadsize && srcPos.x >= -1 * p.roadsize
    && dstPos.x <= p.roadsize && dstPos.x >= -1 * p.roadsize)
    los = true;
  if (srcPos.y <= p.roadsize && srcPos.y >= -1 * p.roadsize
    && dstPos.y <= p.roadsize && dstPos.y >= -1 * p.roadsize)
    los = true;
  

  if (!los) { 
    // vehicles are not on same road
    // check if they are LOS on perpendicular roads

    // distances to virtualSource = crossingCenter
    dr = sqrt(dstPos.x*dstPos.x + dstPos.y*dstPos.y);
    dt = sqrt(srcPos.x*srcPos.x + srcPos.y*srcPos.y);
    
    // sender distance to wall
    if (srcPos.y >= -1 * p.roadsize && srcPos.y <= p.roadsize) {  // sender is on road y=~0
      xt = (dstPos.y >= 0) ? p.roadsize - srcPos.y : srcPos.y + p.roadsize;
    } else {
      xt = (dstPos.x >= 0) ? p.roadsize - srcPos.x : srcPos.x + p.roadsize;
    }

    // check if vehicles are in LOS
    // this check assumes that both vehicles are on perpendicular roads
    double losDr = 0; // the receiver distance to virtual source where NLOS->LOS
    losDr = dt * (xt / (dt - p.roadsize));  // TODO: this calculation might not be 100% correct
        // as dt is distance to center and not parallel to wall
    if (dr <= losDr) {
      los = true;
    }
  }

  
  if (!los) {
    // NLOS rx power calculation according to formula (p. 78)
    // done in dBm space!
    Pr0 = getPr0 (dr, dt, xt, p.roadsize, p.suburban);
    nlosPowerDbm = p.txPowerDbm - p.systemloss - Pr0;
  
    // model NLOS fading with a normal distribution with sigma = 4.1dB   
    rxPowerDbm = nlosPowerDbm;
    
  } else {
      // simply use LOS, no combination of LOS and NLOS!
      
    double plRef = 47.86; // reference loss at 1m
    Vector vdist = vdiff(srcPos, dstPos);
    double dist = sqrt(vdist.x*vdist.x + vdist.y*vdist.y);
    
    if (dist > 1) {
      losPrDbm = p.txPowerDbm - p.systemloss - plRef - 10 * 2.04 * log10(dist);
    } else {
      losPrDbm = p.txPowerDbm - p.systemloss - plRef;
    }
    rxPowerDbm = losPrDbm;
    
    // model LOS fading with a NAK-1 distribution with std-dev = 4.1dB  
  }

  return rxPowerDbm;

}

/* NLOS rx power calculation according to formula (p. 78)
 * done in dBm space!
 */
double
VtlCalc::getPr0 (double dr, double dt, double xt, double roadsize, bool isSuburban) const
{
  const double PI = 3.14159265358979323846;
  double d_b = 4*p.height*p.height / p.lambda;
  float exp_dt = 0.957;
  float exp_st = 0.81;
  double suburban_w = isSuburban ? 2.94 : 0.0;
  double Pr0_t = 0.0;
  
  if (dr <= d_b) {
    Pr0_t = pow(dt,exp_dt) / ((double)pow(xt * 2 * roadsize,exp_st)) *
      (4*PI*dr)/((double)p.lambda);
  } else {
    Pr0_t = pow(dt,exp_dt) / ((double)pow(xt * 2 * roadsize,exp_st)) *
      (4*PI*dr*dr)/((double)p.lambda*d_b);    // dr^2 and lambda*d_b if dr > db
  }
  
  return 3.75 + suburban_w + 10*log10 ((double)pow(Pr0_t, 2.69));
}


/*
 * http://www.johndcook.com/cpp_phi.html
 */
inline double
VtlCalc::Phi (double x, double stddev)
{
  // constants
  double a1 =  0.254829592;
  double a2 = -0.284496736;
  double a3 =  1.421413741;
  double a4 = -1.453152027;
  double a5 =  1.061405429;
  double p  =  0.3275911;

  x = x / stddev; // --Till 

  // Save the sign of x
  int sign = 1;
  if (x < 0)
      sign = -1;
  x = fabs(x)/sqrt(2.0);

  // A&S formula 7.1.26
  double t = 1.0/(1.0 + p*x);
  double y = 1.0 - (((((a5*t + a4)*t) + a3)*t + a2)*t + a1)*t*exp(-x*x);

  return 0.5*(1.0 + sign*y);
}

Vector
VtlCalc::vdiff (const Vector &v, const Vector &w) const
{
  return Vector (v.x - w.x, v.y - w.y, v.z - w.z);
}


} //ns3
