/* -*-  Mode: C++; c-file-style: "gnu"; indent-tabs-mode:nil; -*- */
//
// VTL

#ifndef __vtl_application_h__
#define __vtl_application_h__

#include <map>
#include "ns3/address.h"
#include "ns3/application.h"
#include "ns3/event-id.h"
#include "ns3/ptr.h"
#include "ns3/data-rate.h"
#include "ns3/random-variable.h"
#include "ns3/traced-callback.h"
#include "ns3/header.h"
#include "ns3/tag.h"
#include "ns3/buffer.h"
#include "ns3/vector.h"
#include "ns3/double.h"
#include <ns3/mobility-model.h>
#include <ns3/intelligent-driver-mobility-model.h>
#include "ns3/log.h"
#include "vtl-calc.h"

#define VTL_INT_INVALID 999999999
#define VTL_BYTE_INVALID 255

#define VTL_MIN_SECONDS_DBRAKE 5


#define VTL_HEADER_MAXENTRIES 5
#define VTL_HEADER_MAXROADS 4



#define VTL_LOG_TRACE

#ifdef VTL_LOG_TRACE

#define VTL_LOG_DEBUG(msg) \
  do {                      \
    std::stringstream ss;             \
    ss << msg;    \
    m_vtlDebugTrace (Simulator::Now (), m_id, m_pos, m_dir, m_speed, m_state, UNKNOWN, ss.str()); \
  } while (false)
  
#else

#define VTL_LOG_DEBUG(msg) \
  NS_LOG (ns3::LOG_DEBUG, "#\t" << Simulator::Now().GetSeconds() << "\t" << m_id << "\t" << m_pos << "\t" << msg)

#endif


namespace ns3 {

class Address;
class RandomVariable;
class Socket;

class VtlHeader : public Header
{
public:
  struct VtlInfo {
    Time      validFrom;
    Time      validUntil;
    uint8_t   lightmap[VTL_HEADER_MAXROADS]; // roadlocalid -> red/green
  };
  

  static TypeId GetTypeId (void);
  VtlHeader ();
  virtual ~VtlHeader ();
  virtual TypeId GetInstanceTypeId (void) const;
  virtual uint32_t GetSerializedSize (void) const;
  virtual void Serialize (Buffer::Iterator start) const;
  virtual uint32_t Deserialize (Buffer::Iterator start);
  virtual void Print (std::ostream &os) const;
  
// normal beacon content
  uint32_t            m_source;
  uint32_t            m_seqNo;
  Time                m_timestamp;
  Vector              m_pos;
  double              m_speed;
  
// VTL related data

  uint32_t            m_intersectionId;
  uint8_t             m_localId;
  
  bool                m_isClusterLeader;
  Time                m_clValidity;
  bool                m_isLeader;
  Time                m_leaderValidity;
  bool                m_electionReady;
  Time                m_erValidity;


  uint32_t            m_nextLeader;
  uint32_t            m_currentPhase;
  uint8_t             m_clusterSize;
  double              m_dist;
  
  uint8_t             m_nVtlInfo;
  VtlInfo             m_vtlInfo[VTL_HEADER_MAXENTRIES];
};




/**
 * \ingroup applications
 * \defgroup vtl PeriodicBroadcastApplication
 *
 * This traffic generator follows the periodic broadcast agent style as it is
 * envisioned in vehicular networks.
 *
 * Once the application has been started, it sends packets at a fixed rate
 * , e.g. at 5 Hz
 *
 * The payload of the packets can either be set manually using the Attribute
 * "Payload" or left empty. In the latter case (default), no payload will be used.
 */
class VtlApplication : public Application
{
public:
  static TypeId GetTypeId (void);

  VtlApplication ();
  virtual ~VtlApplication();
  
  
  enum ApproachState {
    FAILSAFE,
    VALIDVTL,
    VALIDVTLDECIDED,
    GREEN,    // only used for output!!!
    RED,      // only used for output!!!
    UNKNOWN   // only used for output in debug messages
  };
  
  enum Role {
    SLAVE,
    CL,
    LEADER,
    LEADER_HANDOVER,
    FINISHED
  };

  enum LeaderStrategy {
    DISTANCE,
    DISTANCESAFE,
    ID,
    PROBABILISTIC
  } e_leaderStrat;

  typedef struct VtlHeader::VtlInfo VtlInfo;
protected:

  virtual void DoDispose (void);
  
private:
  
  
  struct VehicleInfo {
    uint32_t  id;
    Time      timestamp;
    Vector    pos;
    double    speed;
    uint8_t   localId;
    double    dist;
    uint8_t   pgr;  // in Hz
    uint8_t   clusterSize;
    
    bool      isCL;
    Time      clValidity;
    bool      isLeader;
    Time      leaderValidity;
    bool      isER;
    Time      erValidity;
  };
    

    // inherited from Application base class.
  virtual void StartApplication (void);    // Called at time specified by Start
  virtual void StopApplication (void);     // Called at time specified by Stop

    // helpers
  void            CancelEvents ();
  bool            IsConflicting (uint8_t);
  double          GetLeaderRank (double, uint8_t);
  void            InitVtl ();
  void            SetState (Role);
  std::string     GetStateName (Role);
  void            UpdatePositionAndSpeed (void);
  void            StoreInfo (VehicleInfo &vi, const VtlHeader &h);
  void            UpdateVtlInfo (const VtlHeader &h);
  inline bool     HasGreen (void);
  inline bool     HasYellow (void);
  inline bool     AfterDBrake (void);
  bool            CheckBecomeLeader (void);
  bool            CheckBecomeLeaderById (void);
  bool            CheckBecomeLeaderByDistance (void);
  bool            CheckBecomeLeaderByDistanceSafe (void);
  bool            CheckBecomeLeaderProbabilistic (void);
  uint32_t        CheckHandoverLeader (const VtlInfo &vi);
  void            Reset (void);
  void            ResetOtherCL (void);
  Time            calcClValidity(void);
  inline double   DSafe(double speed);
  inline double   GetRequiredDeceleration(double speed, double distance);
  Time            CalcAllYellowInterval (void);
  uint8_t         GetFirstEtaRoad (void);
  uint32_t        FindRoadInRoadMapByLocalId(uint8_t localId);
  bool            RequireAllYellowInterval (void);
  
  
    // Event handlers
  void ReceivePacket (Ptr<Socket> socket);
  void UpdateRoad (uint32_t, uint8_t);
  void NextTimestep();
  void SendPacket();
  void ScheduleNextTx();
  void VtlExpire (void);
  void CheckCurrentVtl (void);
  void LeaderUpdateTimeslice (void);
  
  
  
  // constant/configuration
  Ptr<Socket>   m_socket;     // associated socket
  Address       m_peer;       // peer address
  Time          m_interval;   // the time interval
  TypeId        m_protocol;   // the underlying protocol which is used
  uint32_t      m_seqNo;      // sequence number counter
  uint32_t      m_id;         // nodeID
  uint32_t      m_packetSize;
  Time          m_clusterInterval;
  bool          m_disabled;
  double        m_maxAcc;
  double        m_safeAcc;
  double        m_reactionTime;
  std::string   m_leaderStrategy;
  Time          m_leaderTimeslicePretime;
  
  // pointer to idm, navi
  Ptr<IntelligentDriverMobilityModel> m_mob;   // mobility model
  Ptr<Navigator>                      m_navi;
  Ptr<VtlCalc>                        m_vtlcalc;

  // position information
  uint32_t      m_intersectionId;
  uint8_t       m_localId;        // local id
  Vector        m_pos;        // current position
  Vector        m_dir;        // current direction
  double        m_speed;      // current speed
  double        m_dist;
  double        m_acc;
  
  
  // VTL state
  ApproachState m_approach;
  Role          m_state;
  uint32_t      m_handoverTo;
  
  
  // information store
  VehicleInfo                       m_clmap[VTL_HEADER_MAXROADS];
  std::list<VtlInfo>                m_vtlList;    
  Time                              m_foundBetterClusterLeaderTime;  
  std::map<uint32_t,VehicleInfo>    m_ownCluster;
  Time                              m_clObservationStart;
  
  // traffic light management
  Time          m_minPhaseLength;
  Time          m_maxPhaseLength;
  uint32_t      m_currentPhase;
  std::vector<std::list<Ptr<Road> > > m_roadmap;
  
  
  // helper, events etc.
  EventId       m_sendEvent;
  EventId       m_timestepEvent;
  EventId       m_leaderUpdateEvent; 
  EventId       m_leaderHandoverDegradeEvent; 
  
  UniformVariable jitterGenerator;

  TracedCallback<Time> m_clTrace;
  TracedCallback<Time> m_clConflictTrace;
  
  TracedCallback<Time> m_vtllElectionTrace;
  TracedCallback<Time> m_vtllHandoverTrace;
  TracedCallback<Time,double> m_vtllExceptionTrace;
  
  TracedCallback<Time> m_leaderConflictTrace;
  TracedCallback<Time,double> m_finishedTrace;
  
  TracedCallback<Time,double> m_requiredDecelerationTrace;
  
  
  TracedCallback<Time,uint32_t,Vector,Vector,double,Role,ApproachState> m_vtlTrace;
  TracedCallback<Time,uint32_t,Vector,Vector,double,Role,ApproachState,std::string> m_vtlDebugTrace;
  
  
  
  
};

} // namespace ns3

#endif

