/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * 
 *
 */
 
 
#include "ns3/log.h"
#include "manhattan-idm.h"
#include <math.h>

NS_LOG_COMPONENT_DEFINE ("ManhattanIDM");
namespace ns3 {



////////////////////////
// IDMUtils
////////////////////////

ns3::Vector
IDMUtils::vmult(const Vector &v, double d)
{
  return Vector(v.x * d, v.y * d, v.z * d);
}

ns3::Vector
IDMUtils::vplus(const Vector &v, const Vector &w)
{
  return Vector(v.x + w.x, v.y + w.y, v.z + w.z);
}

ns3::Vector
IDMUtils::vdiff(const Vector &v, const Vector &w)
{
  return Vector(v.x - w.x, v.y - w.y, v.z - w.z);
}

double 
IDMUtils::vlength(const Vector &v)
{
  double length = sqrt (v.x * v.x + v.y * v.y + v.z * v.z);
  return length;
}

double 
IDMUtils::vdist (const Vector &a, const Vector &b)
{
  double dx = b.x - a.x;
  double dy = b.y - a.y;
  double dz = b.y - a.z;
  
  double distance = sqrt (dx * dx + dy * dy + dz * dz);
  return distance;
}

ns3::Vector
IDMUtils::vnorm(const Vector &v)
{
  return vmult(v, 1.0/vlength(v));
}



////////////////////////
// Intersection
////////////////////////

NS_OBJECT_ENSURE_REGISTERED (Intersection);

TypeId Intersection::GetTypeId (void) {
  static TypeId tid = TypeId ("ns3::Intersection")
    .SetParent<Object> ()
    .AddConstructor<Intersection> ();
  return tid;
}

Intersection::Intersection () {
  m_id = IDM_INT_INVALID;
  m_ibd = 20;
  m_suburban = false;
  m_stopDist = 5;
  m_localIdCount = 0;
  m_lastAdded = NULL;
  m_lastAdded2 = NULL;
}

Intersection::~Intersection () {

}

std::vector<std::list<Ptr<Road> > >
Intersection::GetRoadMap (void)
{
  if (m_rl.size () == 0)
    MakeRoadMap ();
  return m_rl;
}


/*
 * mark two roads as non conflicting so that vehicles do not wait
 * for vehicles on other road
 */
void
Intersection::AddNoConflict (Ptr<Road> r1, Ptr<Road> r2)
{
  NS_LOG_DEBUG("setting noconflict at intersection " << m_id << " for roads: " << r1->m_id << " and " << r2->m_id);
  m_noconflict[r1->m_id].push_back (r2);
  m_noconflict[r2->m_id].push_back (r1);
}

/*
 * sets no conflict on the two last added roads
 */
void
Intersection::AddNoConflict()
{
  NS_ASSERT (m_lastAdded != NULL && m_lastAdded2 != NULL);
  
  AddNoConflict (m_lastAdded, m_lastAdded2);
}

/*
 * checks if two roads are conflicting
 */
bool
Intersection::IsRoadConflicting (uint32_t rid1, uint32_t rid2)
{
  std::list<Ptr<Road> > nc;
  std::list<Ptr<Road> >::iterator it;
  
  nc = m_noconflict[rid1];
  for (it=nc.begin(); it != nc.end(); ++it) {
    if ((*it)->m_id == rid2) {
      return false; // rid2 was in rid1's non-conflicting set
    }
  }
  
  return true;  // roads conflict by default
}

bool
Intersection::CanGo (Ptr<Road> road, uint32_t vid) {
  queueEntry first = m_queues[road->m_prio].front();
  if (first.vid == vid) {
    return true;
  } else if (!IsRoadConflicting(road->m_id, first.rid)) {
    // move to head of queue in order to prevent conflicting vehicles from
    // entering the intersection too early (i.e., when the first vehicle already
    // crossed intersection but not me)
    // moving two second position would be enough but first is ok as well
    RemoveQueued(road, vid);
    AddToQueueFront(road, vid);
    return true;
  }
  return false;
  
}

void
Intersection::AddToQueue (Ptr<Road> road, uint32_t vid) {
  uint8_t prio = road->m_prio;
  uint32_t rid = road->m_id;
  NS_LOG_DEBUG("Intersection::AddToQueue at intersection " << m_id << " for road " << rid << " and vehicle " << vid);
  queueEntry e;
  e.vid = vid;
  e.rid = rid;
  m_queues[prio].push_back (e);
}

void
Intersection::AddToQueueFront (Ptr<Road> road, uint32_t vid) {
  uint8_t prio = road->m_prio;
  uint32_t rid = road->m_id;
  NS_LOG_DEBUG("Intersection::AddToQueueFront at intersection " << m_id << " for road " << rid << " and vehicle " << vid);
  queueEntry e;
  e.vid = vid;
  e.rid = rid;
  m_queues[prio].push_front (e);
}

bool
Intersection::RemoveQueued (Ptr<Road> road, uint32_t vid) {
  std::list<queueEntry>::iterator it;
  uint8_t prio = road->m_prio;
  
  for (it=m_queues[prio].begin(); it != m_queues[prio].end(); ++it) {
    if (it->vid == vid) {
      m_queues[prio].erase(it);
      NS_LOG_DEBUG("Intersection::RemoveQueued at intersection " << m_id << " for vehicle " << vid << " success");
      //PrintQueues ();
      return true;      
    }
  }
  NS_LOG_DEBUG("Intersection::RemoveQueued at intersection " << m_id << " for vehicle " << vid << " FAILED");
  //PrintQueues ();
  return false;
}

void
Intersection::PrintQueues (void)
{
  NS_LOG_DEBUG ("PrintQueues at intersection "<< m_id);
  std::map<uint8_t,std::list<queueEntry> >::iterator qit;
  std::list<queueEntry>::iterator lit;
  for (qit = m_queues.begin(); qit != m_queues.end(); ++qit) {
    NS_LOG_DEBUG ("PRIO = " << (int) qit->first);
    for (lit = qit->second.begin(); lit != qit->second.end(); ++lit) {
      NS_LOG_DEBUG ("vehicle = " << lit->vid << " road = " << lit->rid);
    }
  }
}

bool
Intersection::CompareRoadPrio (Ptr<Road> r1, Ptr<Road> r2) {
  return r1->m_prio > r2->m_prio;
}


void
Intersection::AddApproachRoad (Ptr<Road> road)
{
  m_lastAdded2 = m_lastAdded;
  m_lastAdded = road;
  
  m_roadsTo.push_back (road);
  m_localIdToRoad[m_localIdCount] = road;
  road->m_localId = m_localIdCount;
  m_localIdCount++;
}

void
Intersection::MakeRoadMap ()
{
  uint32_t i = 0;
  bool foundConflict;
  
  std::list<Ptr<Road> >::iterator it;
  std::list<Ptr<Road> >::iterator rit;
  
  m_roadsTo.sort (Intersection::CompareRoadPrio);
  m_rl.clear();
  NS_LOG_DEBUG("MAKEROADMAP");
  for (it = m_roadsTo.begin(); it != m_roadsTo.end(); ++it) {
    NS_LOG_DEBUG("road id =" << (*it)->m_id << " prio=" << (int)(*it)->m_prio);
    i = 0;
    while (true) {
      if (i >= m_rl.size ()) {
        // empty set, add here in new list!
        std::list<Ptr<Road> > tlist;
        tlist.push_back (*it);
        m_rl.push_back(tlist);
        break;      
      }
      foundConflict = false;
      for (rit = m_rl[i].begin(); rit != m_rl[i].end(); ++rit) {
        if (IsRoadConflicting((*it)->m_id, (*rit)->m_id)) {
          foundConflict = true;
          break;
        }
      }
      if (!foundConflict) {
        // no conflict found in this set, add here!
        m_rl[i].push_back (*it);
        break;
      }
      i++;
    }
  }
}

inline Ptr<Road>   
Intersection::GetRoadByLocalId(uint8_t lid){
  return m_localIdToRoad[lid];
}

////////////////////////
// Road
////////////////////////


std::vector<Ptr<Road> > Road::roads;

NS_OBJECT_ENSURE_REGISTERED (Road);

TypeId Road::GetTypeId (void) {
  static TypeId tid = TypeId ("ns3::Road")
    .SetParent<Object> ()
    .AddConstructor<Road> ();
  return tid;
}

Road::Road ()
  : m_vehicles () {
  m_prio = 0;

}

Road::~Road () {
}

void
Road::addVehicle(Ptr<Navigator> navi)
{
  //NS_LOG_DEBUG("push vehicle to road");
  m_vehicles.push(navi);
}

/*
 * returns time until next vehicle reaches intersection
 */
double
Road::GetGapTime()
{
  if (m_vehicles.size() == 0) {
    return 999;
  } else {
    Ptr<Navigator> next = m_vehicles.front();
    return next->GetTimeToIntersection();
  }
}

/*
 * adds all roads at next intersection with higher priorities
 * to m_roadsHiPrio. This is used later by navigator to give way to
 * other vehicles on these roads
 * called by topology maker after all roads and intersections are connected
 * (could also retrieve some information used for routing in the future)
 */
void
Road::CheckNextIntersection () {

  std::list<Ptr<Road> > roadsTo;
  
  if (m_prio == IDM_HIGH_PRIO)
    return; // highest prio, always right of way
    
  if (m_end == NULL)
    return; // road is sink road, no intersection

  roadsTo = m_end->m_roadsTo;
    
  m_roadsHiPrio.clear ();
  
  std::list<Ptr<Road> >::iterator it;
  for (it = roadsTo.begin(); it != roadsTo.end(); ++it) {
    if ((*it)->m_prio > m_prio) { // other has higher prio
      m_roadsHiPrio.push_back (*it);
    }
  }
}

Ptr<Road>
Road::GetHighPrioRoad ()
{
  std::vector<Ptr<Road> >::iterator it;
  Ptr<Road> highroad = NULL;
  
  for (it = m_roadsHiPrio.begin(); it != m_roadsHiPrio.end(); ++it) {
    if (highroad == NULL || (*it)->m_prio > highroad->m_prio) {
      highroad = *it;
    }
  }
  return highroad;
}

double
Road::GetHighRoadGap () {
  double minGap = 999;
  double gap;
  std::vector<Ptr<Road> >::iterator it;
  for (it = m_roadsHiPrio.begin(); it != m_roadsHiPrio.end(); ++it) {
    gap = (*it)->GetGapTime();
    if (gap < minGap)
      minGap = gap;
  }
  return minGap;
}

Road::roadrelation
Road::GetRelationTo (Ptr<Road> other)
{
  if (other->m_dir.x == -1*m_dir.x && other->m_dir.y == -1*m_dir.y) {
    return LOS;
  } else if (other->m_dir.x == m_dir.y 
              && other->m_dir.y == -1*m_dir.x) {
    return LEFT;
  } else if (other->m_dir.x == -1*m_dir.y 
              && other->m_dir.y == m_dir.x) { 
    return RIGHT;
  } else {
    NS_ASSERT(false);
    return LOS;
  }
}

////////////////////////
// Navigator
////////////////////////

NS_OBJECT_ENSURE_REGISTERED (Navigator);

TypeId Navigator::GetTypeId (void) {
  static TypeId tid = TypeId ("ns3::Navigator")
    .SetParent<Object> ()
    .AddConstructor<Navigator> ();
  return tid;
}

Navigator::Navigator () {
  m_astate = INIT;
}

Navigator::Navigator (Ptr<Road> source) {
  m_curRoad = source;
  m_astate = INIT;
  m_running = false;
}

Navigator::~Navigator () {

} 


Navigator::RoadState 
Navigator::CheckRoadChange (double distDriven) {
  if (m_astate == CROSSING2 && distDriven > m_lastInt->m_stopDist + m_carLength/2.0) {
    bool ret = m_lastInt->RemoveQueued (m_lastRoad, m_id);
    NS_ASSERT (ret);
    m_astate = INIT;
  }

  if (m_curRoad->m_length < distDriven) {
    // we left old street
    m_curRoad->m_vehicles.pop ();
    m_idm->m_distDriven = 0;  // IDM sets this on return of this function, but application may
                              // call GetStopDist on roadchange and get wrong values
    
    // after crossing an intersection in failsafe mode, schedule remove vehicle
    // from fifo queue in +1 second as vehicle is now in the middle of the intersection
    if (m_astate == CROSSING) {
      if (m_curInt != NULL && m_curRoad->m_prio != IDM_HIGH_PRIO) {
        m_astate = CROSSING2;
        m_lastInt = m_curInt;
        m_lastRoad = m_curRoad;
      } else {
        NS_LOG_DEBUG ("Crossed intersection with high prio");
        m_astate = INIT;
      }
    } else {
      NS_LOG_DEBUG ("Crossed intersection in state " << m_astate);
      m_astate = INIT;
    }
    
    if (m_curRoad->m_end == NULL) {
      // vehicle reached end of sink road
      m_curRoad = NULL;
      m_running = false;
      return FINISHED;
    } else {
      // choose next road, could be some routing stuff here
      UpdateRoad (Road::roads[m_curRoad->m_id + 1]);
      return CHANGED;
    }

  } else {
    return UNCHANGED;
  }
}



Vector
Navigator::GetCurrentDirection () {
  return m_curRoad->m_dir;
}

Vector
Navigator::GetCurrentRoadStart () {
  return m_curRoad->m_startPos;
}

Ptr<IntelligentDriverMobilityModel>
Navigator::GetCurrentPOV () {
  if (m_pov)
    return m_pov->m_idm;
  return NULL;
}

double
Navigator::GetStopDist ()
{
  if (m_curInt)
    return m_curRoad->m_length - m_idm->m_distDriven - m_curInt->m_stopDist;
  return 999; 
}


double
Navigator::GetCompleteCrossingDist ()
{
  if (m_curInt)
    return m_curRoad->m_length - m_idm->m_distDriven + m_curInt->m_stopDist + m_carLength/2.0;
  return 999; 
}

Road::roadrelation
Navigator::GetRelationTo (uint8_t otherid)
{
  Ptr<Road> otherp = m_curInt->GetRoadByLocalId(otherid);
  return m_curRoad->GetRelationTo(otherp);
}

/*
 * returns the estimated time required to reach the intersection center 
 * at the current speed
 * TODO use speed AND acceleration
 * s = v0*t + 1/2 a*t^2
 * t = (-v  +- sqrt(2a*s+v^2)) / a
 */ 
double
Navigator::GetTimeToIntersection ()
{
  double speed = m_idm->GetSpeed ();
  double dist = GetStopDist ();
  if (speed > 0) {
    return dist / speed;
  } else {
    return 999;
  }
}

bool
Navigator::IsFirstVehicle ()
{
  return m_curRoad->m_vehicles.front() == this;
}

/*
 * checks if a vehicle can cross the current intersection by checking if
 * 1. we are first of same prio FCFS queue
 * 2. the gap on all higher prio roads is large enough for us to cross
 */
bool
Navigator::CheckGo (double requiredGap) {
  bool cango = m_curInt->CanGo(m_curRoad, m_id);
  if (cango) {
    // we are first vehicle, check higher prio roads for gap  
    return (requiredGap < m_curRoad->GetHighRoadGap ());
  } else {
    return false; // wait for other, same prio vehicle to leave intersection
  }
}

/*
 * performs failsafe intersection crossing
 * if vehicle is on IDM_HIGH_PRIO prio road, vehicle does not stop
 * if vehicle is on normal road, vehicle stops at intersection,
 * gives right of way to higher prio roads and performs FCFS with
 * same prio roads.
 * calls idm->stopat and idm->resume for stop and resume
 * 
 */
void
Navigator::checkFailsafe (double speed, double requiredGap)
{
  if (m_astate == INIT) {
    // set stoppoint only if not on IDM_HIGH_PRIO prio road
    if (m_curRoad->m_prio != IDM_HIGH_PRIO && m_curInt != NULL) {
      m_idm->StopAt (GetStopDist ());
      m_astate = APPROACH;
    } else {
      m_idm->Resume ();
      m_astate = CROSSING;
    }
  } else if (m_astate == APPROACH) {
    // check if vehicle stopped at stop line
    bool stopped = (speed < 0.1 && IsFirstVehicle() && GetStopDist() < 5.0);
    if (stopped) {
      m_astate = WAITING;
      m_curInt->AddToQueue (m_curRoad, m_id);
    }
  } else if (m_astate == WAITING) {
    if (CheckGo (requiredGap)) {
      m_idm->Resume ();
      m_astate = CROSSING;
    }
  }
}

/*
 * failsafe approach mode can be ended by the application at any time
 * therefore, any leftover must be removed.
 * removes vehicle from waiting queue at intersection and resets state to init
 */
void
Navigator::EndFailsafe ()
{
  if (m_astate == WAITING 
      || (m_astate == CROSSING  && m_curRoad->m_prio != IDM_HIGH_PRIO)) {
    m_curInt->RemoveQueued (m_curRoad, m_id);
  } else if (m_astate == CROSSING2 && m_lastRoad->m_prio != IDM_HIGH_PRIO) {
    m_lastInt->RemoveQueued (m_lastRoad, m_id);
  }
  
  m_astate = INIT;
}

/*
 * invoked by idm, adds vehicle to source road and sets
 * POV vehicle (vehicle directly in front) if any
 */
void
Navigator::Start() {
  m_running = true;
  m_carLength = m_idm->GetCarLength();
  UpdateRoad (m_curRoad);
}

void
Navigator::UpdateRoad (Ptr<Road> r) {
    
  m_curRoad = r;
  m_curInt = m_curRoad->m_end;
  if (m_curRoad->m_vehicles.size() > 0) {
    m_pov = m_curRoad->m_vehicles.back();
  } else {
    m_pov = NULL;
  }
  m_curRoad->m_vehicles.push (this);
  if (!m_roadUpdate.IsNull()) {
    m_roadUpdate (GetCurrentIntersectionId(), GetCurrentLocalId());
  }
}

/*
 * Checks if the vehicle is still running and returns the common intersection
 * of two vehicles, if both vehicles are on roads that are attached to one intersection
 * returns null if any vehicle already finished or vehicles are not on adajcent
 * roads. 
 */
Ptr<Intersection>
Navigator::GetCommonIntersection (Ptr<Navigator> navi)
{
  if (m_running == false)
    return NULL;  // vehicle finished
  if (navi->m_running == false)
    return NULL;  // other vehicle finished
  
  Ptr<Road> road = navi->m_curRoad;
  if (m_curRoad->m_start != NULL 
      && (m_curRoad->m_start == road->m_start || m_curRoad->m_start == road->m_end))
    return m_curRoad->m_start;
  if (m_curRoad->m_end != NULL 
      && (m_curRoad->m_end == road->m_start || m_curRoad->m_end == road->m_end))
    return m_curRoad->m_end;    
    
  return NULL;  // no common intersection found
}

uint32_t
Navigator::GetCurrentIntersectionId (void)
{
  if (m_curInt != NULL)
    return m_curInt->m_id;
  return IDM_INT_INVALID;
}

uint8_t
Navigator::GetCurrentLocalId (void)
{
  if (m_curRoad != NULL)
    return m_curRoad->m_localId;
  return IDM_BYTE_INVALID;
}

void
Navigator::SetUpdateCallback (Callback<void, uint32_t, uint8_t> update)
{
  m_roadUpdate = update;
}

} // ns3 namespace
